/*
lfmtix -- reformat text data.
    Copyright (C) 2003  Norm London, London Computing

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	norm_london@londoncomputing.com

    Compile with gcc:  gcc lfmtix.c -o lfmtix 
 
*/


#define MAIN
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <fcntl.h>
#include <stddef.h>
#include <sys/types.h>
#include <sys/stat.h>


#ifdef WIN32
#include <dos.h>
#include <windows.h>
#include <io.h>
#include <process.h>
#else
#define _O_BINARY
#define _MAX_PATH 1024
#define stricmp strcasecmp
#endif

#define MAXLINE 4096
#define OUTLEN  4096

#define MAXTOK 255
char tokbuf[MAXLINE+64];

FILE *infi, *otfi;
int lineno = 0, usenewline = 1;
char inbuff[8192],outbuff[8192];
char cmdchar='%';
char line[MAXLINE+64];
char format[255];
int done=0; //set to 1 if eof reached unexpectedly
extern int _fmode;

// Modified 9/8/99 to fix tokenization.  Using strtok
// Modified 9/9/99 to remove strtok and use my_strtok instead
// Modified 9/13/99 to add $fn- which tells the program to take
// all of the text from field n to the end of the record and 
// write it to the output
// Modified 9/14/99 to work in binary mode so that char(26)
// are ignored.  This is now a 32 bit application.
// Modified 11/21/99 to fix the previous delim fix.  Needed special processing 
// when one of the delimiters is a space, forcing leading blanks to be 
// removed before looking for the end of field.  For all other delimiters
// this should not be done so that the program will handle empty delimited
// fields, e.g. ,,

char *trim(char *buf) {
	char *b;
	int ln,c;

	b=buf;
	// remove white space from start and end of buf
	if (buf == NULL)
		return NULL;
	// remove leading white space
	while (isspace(*b))
		++b;
	// if string is empty, return pointer to empty string
	if (*b == 0)
		return b;
	// remove trailing white space
	for (ln = strlen(b)-1; ln >= 0; --ln) {
		c = *(b+ln);
		if (isspace(c))
			*(b+ln) = 0;
		else
			return b;
	}
	return NULL;		

}

char *my_strtok(char *str, char *delim) {
	static char *starthere;
	char *s, *start;
	char *d=delim;
	char prev_delim;
	char *match_cur = NULL;
	char *match_min = NULL;

	if (str) 
		// If a new string, set starting postion.
		starthere = str;

	// start at starting position
	s = starthere;

	if (strchr(delim,' ')) {
		// One of the delimiters is a space character
		// remove whitespace at the beginning of the string
		while (*s==' ')
			++s;
	}

	if (*s == 0)
		return NULL;

	start = s;

	// Look for 1st matching delim character
	while (*d) {
		if (match_cur = strchr(s, *d++)) {
			// If *d is a space character, then look for 1st non space
			if (match_min == NULL)
				match_min = match_cur;
			if (match_min > match_cur)
				match_min = match_cur;
		}
	}

	// done scanning
	if (match_min) {
		// found a match; terminate string
		prev_delim = *match_min;
		*match_min = 0;
		++match_min;
		starthere = match_min;
		if (prev_delim == ' ' && strchr(delim,' ')) {
			// a space character is one of the delimiters
			// when this happens, look for the next non space
			// If this character is also a delim then set starthere
			// at the position past this non blank delimiter
			// otherwise use the first space as the delim
			while (*match_min == ' ')
				++match_min;
			if (*match_min && strchr(delim,*match_min))
				starthere = match_min + 1;
		}
	}
	else 
		starthere = start + strlen(start);

	return start; 
}

void get_tokens(char *buf, char *token[], char *delim) {
	int i;
	char *tok,*b;

	b = tokbuf;
	strcpy(tokbuf,buf); // Move into token buffer
	// look for a delim character, using my_strtok
	// strtok can not be used because it removes
	// multiple occurrences of delimiters concatenated
	// together
	tok = my_strtok(b,delim);
	for (i=0; tok && i < MAXTOK;++i) {
		token[i] = trim(tok);
		tok = my_strtok(NULL,delim); // other times
	}
	if (i < (MAXTOK -1))
		token[++i] = NULL;
}
		

void print_token(char *token, int offset, int len, int width, 
		 char rightjustify, FILE *otfi) {
	// Write the token to otfi
	if (offset >= len)  {
		/* don't print anything if offset past end of line */
		// If a field width was specified, fill with blanks
		while(width--)
			putc(' ',otfi);
		return;
	}
	
	if (width == 0) {
		width = len;
		if (len < (width+offset))
			width = len - offset;
		fwrite(token+offset,width,1,otfi);
		return;
	}

	// If the field length is >= w,just write the field data
	// truncate at width
	if ((len - offset) >= width)
		fwrite(token+offset,width,1,otfi);
	else {
		// write out as much data as is present, then pad
		if (rightjustify) {
		// pad first
			width -= (len - offset);
			while(width--)
				putc(' ',otfi);
		}
		// write the data
		fwrite(token+offset,len-offset,1,otfi);
		if (rightjustify == 0) {
			// pad after
			width -= (len - offset);
			while(width--)
				putc(' ',otfi);
		}
	}
	return;
}


void expand(char *s,char *l1, char *delim) {
	int w,o,tok,last_token;
	int flen,len,expanding;
	char *token[MAXTOK+1];
	char tokens = 0;
	char temp[7],*c;
	char *l, rightjustify=0, *start;

	l = l1; // Start using the buffer passed to expand

	if (l[(len = strlen(l))-1] == '\n')
		l[--len] = 0;
	if (l[(len = strlen(l))-1] == '\r')
		l[--len] = 0;
	expanding = 0;

	// Initialize array to all empty strings
	for (tok=0; tok <= MAXTOK; token[tok++] = NULL);

	while (*s) {
		start = s;
		if (!expanding) {
			w = 0;
			o = 0;
			/* now position 1 char past % */
			c = strchr(s,cmdchar);
			if (c == NULL) { /* nothing to expand, just write the format out */
				fputs(s,otfi);
				if (usenewline)
					putc('\n',otfi);
				return;
			}
			if (c > s) { /* write out everything up to % */
				fwrite(s,1,c-s,otfi);
				s = c;
			}
			++s;
			rightjustify = 0;
		}
		expanding = 1;
		switch (toupper(*s)) {
		case 'R':
			if (fgets(line,sizeof(line),infi) == NULL) {
				done = 1;
				if (usenewline) 
					putc('\n',otfi);
				return;
			}
			else 
				l = line;
			// Remove trailing CR and LF
			if (l[(len = strlen(l))-1] == '\n')
				l[--len] = 0;
			if (l[(len = strlen(l))-1] == '\r')
				l[--len] = 0;

			++s;
			o = 0;
			tokens=0;
			expanding = 0;
			break;
		case 'Q':
			putc('\"',otfi);
			++s;
			expanding = 0;
			break;
		case 'S':
			putc('/',otfi);
			++s;
			expanding = 0;
			break;
		case 'N':
			putc('\n',otfi);
			++s;
			expanding = 0;
			break;
		case 'G':
			putc('>',otfi);
			++s;
			expanding = 0;
			break;
		case 'L':
			putc('<',otfi);
			++s;
			expanding = 0;
			break;
		case 'B':
			putc('|',otfi);
			++s;
			expanding = 0;
			break;
		case 'T':
			putc('\t',otfi);
			++s;
			expanding = 0;
			break;
		case '.':
			// If the period follows the control character ($ or %)
			// and if it precedes a digit (0..9), then set the right
			// justify flag, else print out the period.
			if (isdigit(*(s+1))) {
				rightjustify = 1;
				++s;
			}
			else {
				putc('.',otfi);
				expanding = 0;
				++s;
			}
			break;
		case '0':
		case '1':
		case '2':
		case '3':
		case '4':
		case '5':
		case '6':
		case '7':
		case '8':
		case '9':
			w = 10*w + (*s - '0');
			++s;
			break;
		case ' ':
			if (w)
				while (w--)
					putc(' ',otfi);
			else
				putc(' ',otfi);
			++s;
			expanding = 0;
			break;
		case 'F':
			expanding = 0;
			// find first 255 tokens if present
			// if a field width set, then ensure that the field
			// is always the specified size, even if it means
			// padding with blanks.
			if (tokens == 0) {
				get_tokens(l,token, delim);
				tokens = 1;
			}
			tok = 0;
			if (isdigit(*(++s))) {
				while(isdigit(*s)) 
					tok = tok*10 + (*s++ - '0');
				tok --;		// Make token number relative 0, not 1
				s--;		// Make s point to last character in $f
			}
			else {
				if (*s == '=') {
					// This is shorthand for f1=$f1, f2=$f2, ...
					// ignore field width in this case
					for (tok=0;token[tok];++tok) {
						if (tok)
							fprintf(otfi,", ");
						fprintf(otfi,"$f%d=%s",tok+1,token[tok]);
					}
					fprintf(otfi,". ");
					++s;
					break;
				}
				else
					if (*s)
						tok = *s - 'A'+10;
			}

			if (token[tok] == NULL) {
				++s;
				while(w--)
					putc(' ',otfi);
				break; // field does not exist, write empty string
			}
			else
				flen = strlen(token[tok]);
			
			if (*(++s) == '+') {
				++s;
				while (isdigit(*s))
					o = o * 10 + (*s++ - '0');
			}

			last_token = tok;
			if (*s == '-') {
				// specified a range of fields, ie. $f1-15
				// or all the rest of the fields $f1-
				last_token = 0;
				++s;
				while (isdigit(*s))
					last_token = last_token * 10 + (*s++ - '0');
				if (last_token == 0)
					if (isalpha(*s))
						last_token = *s - 'A' + 10;
				if (last_token == 0)
					last_token = MAXTOK;

				if (last_token < tok) {
					*(++s) = 0;
					fprintf(stderr,"Invalid Specification: %s\n",start);
					exit(4);
				}
			}
			
			while(token[tok] && tok <= last_token) {
				print_token(token[tok],o,flen,w,rightjustify,otfi);
				++tok;
			}
			break;

		case 'C':
			expanding = 0;
			if (*(++s) == '+') {
				++s;
				while (isdigit(*s))
					o = o * 10 + (*s++ - '0');
			}
			if (o >= len)  {
				/* don't print anything if offset past end of line */
				// If a field width was specified, fill with blanks
				while(w--)
					putc(' ',otfi);
				break;
			}

			if (w == 0) {
				// If no length specified, put out entire field
				w = len;
				if (len < (w+o))
					w = len - o;
				fwrite(l+o,w,1,otfi);
			}
			else {
				// If the field length is >= w,just write the field data
				if ((len - o) >= w)
					fwrite(l+o,w,1,otfi);
				else {
					// write out as much data as is present, then pad
					if (rightjustify) {
					// pad first
						w -= (len - o);
						while(w--)
							putc(' ',otfi);
					}
					// write the data
					fwrite(token[tok]+o,len-o,1,otfi);
					if (rightjustify == 0) {
						// pad after
						w -= (len - o);
						while(w--)
							putc(' ',otfi);
					}
				}
			}
			
			
			break;

		case '#':
			if (w > 5 || w < 1) w = 5;
			sprintf(temp,"%c%dd",'%',w);
			fprintf(otfi,temp,lineno);
			++s;
			expanding = 0;
			break;
		default:
			putc(*s,otfi);
			expanding = 0;
			++s;
			break;
		}
	}
	if (usenewline) 
		putc('\n',otfi);
	/* bwr(otfi,"\n",2); */
}



void Usage() {
	FILE *hlp;
#ifdef WIN32

	char tmp[_MAX_PATH],*p,cmd[_MAX_PATH];

	if ((p=getenv("tmp")) == NULL)
		if ((p = getenv("temp")) == NULL)
			GetWindowsDirectory(tmp,_MAX_PATH);
		else
			strcpy(tmp,p);
	else
		strcpy(tmp,p);
	if (tmp[strlen(tmp)-1] != '\\')
		strcat(tmp,"\\");

	strcat(tmp,"lfmt32.hlp");
	hlp = fopen(tmp,"w");

#else
	char tmp[1024], *buffer;
	buffer = getenv("HOME");
	if (buffer != NULL) {
		strcpy(tmp,"less ");
		strcat(tmp, buffer);
		if (tmp[strlen(tmp)-1] != '/')
			strcat(tmp,"/");
		strcat(tmp,"temp");
		mkdir(tmp+5,S_IRWXU|S_IRWXG);
		strcat(tmp,"/lfmtix.hlp");
		fprintf(stderr,"%s\n",tmp);
		hlp = fopen(tmp+5,"w");
	}
	if (hlp == NULL)
		hlp = stderr;
#endif
	fprintf(hlp,
	"lfmtix(c) Copyright Norm London, London Computing, 1987,1988,1992,1994,1997-2003 [%s]\n",__DATE__);
	fprintf(hlp,"norm_london@londoncomputing.com\n\n");
        fprintf(hlp,"This program is free software; you can redistribute it and/or modify\n");
	fprintf(hlp,"it under the terms of the GNU General Public License as published by\n");
	fprintf(hlp,"the Free Software Foundation; either version 2 of the License, or\n");
	fprintf(hlp,"(at your option) any later version.\n\n");
	fprintf(hlp,"This program is distributed in the hope that it will be useful,\n");
	fprintf(hlp,"but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
	fprintf(hlp,"MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
	fprintf(hlp,"GNU General Public License for more details.\n\n");
	fprintf(hlp,"You should have received a copy of the GNU General Public License\n");
	fprintf(hlp,"along with this program; if not, write to the Free Software\n");
	fprintf(hlp,"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}
	fprintf(hlp,"Usage is:\nlfmtix [-Q] [-Sn] [-D=\"xxx\"] [{-C | inputfile | NUL}] [/format] [-ooutfile] [-n] [-$]\n");
	fprintf(hlp,"For LINUX. Prints a file according to an output format.   Where:\n");
	fprintf(hlp,"inputfile -- is the name of the file to be formatted\r\n");
	fprintf(hlp,"NUL       -- specifies a file with only one empty record (CR-LF only).\n");
	fprintf(hlp,"             Allows you to use lfmtix to output strings containing special\n");
	fprintf(hlp,"             characters, e.g. < > , | ..\n");
	fprintf(hlp,"-C        -- indicates that the file is to be read from console(STDIN).\n");
	fprintf(hlp,"             -C may be omitted if either a /format or a flag is set\n");
	fprintf(hlp,"             on the command line.\n\n");
	fprintf(hlp,"If neither an inputfile nor -C is specified, read from console(STDIN)\n");	
	fprintf(hlp,"Redirection is permitted.\n\n");
	fprintf(hlp,"-Sn       -- Skip the first n lines of the file.\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}
	fprintf(hlp,"-D=\"xxx\"  -- Specify a set of delimiters to separate arguments.  These\n");
	fprintf(hlp,"             are used to parse the file when a %cFn is used in the format.\n",'%');
	fprintf(hlp,"             %cF1 stands for field 1, ...  Leading and trailing Whitespace \n",'%');
	fprintf(hlp,"             (tab, space) characters are automatically removed.  When\n");
	fprintf(hlp,"             specified any character defined between the quotes will \n");
	fprintf(hlp,"             end a field.  If omitted, a blank or tab is will end a field.\n");
	fprintf(hlp,"             Use \\> for >   \\< for <  \\| for | \\t for tab \n");
	fprintf(hlp,"                  \\n for lf \\r for cr \\\\\\\\ for \\  \\q for \"\n");
	fprintf(hlp,"             A space character is treated differently from all other\n");
	fprintf(hlp,"             delimiter characters.  Multiple spaces are treated as a single.\n");
	fprintf(hlp,"             delimiter and multiple spaces preceding some other delimiter\n");
	fprintf(hlp,"             are also treated as a single delimiter.  When the delimiter\n");
	fprintf(hlp,"             characters is not a space, multiple delimiters are treated\n");
	fprintf(hlp,"             as delimiting separate fields.  For example, if the delimiter\n");
	fprintf(hlp,"             strings is set to %c ,%c (space and comma), then the string\n",'\"','\"');
	fprintf(hlp,"             A  B ,  C will return three fields: F1=A, F2=B and F3=C.\n");
	fprintf(hlp,"             If the delimiter string is set to %c,%c (Comma Only), then the \n");
	fprintf(hlp,"             string A,,B,,C will return 5 fields:\n");
	fprintf(hlp,"             F1=A, F2=(null), F3=B, F4=(null) and F5=C\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}

	fprintf(hlp,"-Q        -- Quiet Mode.  Do not print copyright line.\n");
	fprintf(hlp,"-N        -- if specified, omit newline after processing each record.\n");
	fprintf(hlp,"-%c        -- if specified, use %c instead of \%\n",'%','$');
	fprintf(hlp,"             If you select this option, you will have to double dollar\n");
	fprintf(hlp,"             signs, especially useful within .BAT or .CMD files.\n");
	fprintf(hlp,"-ooutfile -- specifies the name of an output file.  STDOUT is default.\n");
	fprintf(hlp,"format   --  is the format specification for output\r\n");
	fprintf(hlp,"             Default format is: \%5#: \%C\r\n");
	fprintf(hlp,"Format must start with a /.  Format Characters are:\n");
	fprintf(hlp,"\t\%[n]C[+m] --Prints n characters from inputline starting at pos m.\n");
	fprintf(hlp,"\t            n defaults to the length of the line, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the right (Left Justify the field.)\n");
	fprintf(hlp,"\t\%[.n]C[+m]--Prints n characters from inputline starting at pos m.\n");
	fprintf(hlp,"\t            n defaults to the length of the line, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the left (Right Justify the field.)\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}

	fprintf(hlp,"\t\%[n]Fk[+m][-l]--Prints n characters Field k starting at position m.\n");
	fprintf(hlp,"\t            If -l is specified, then fields k through l are printed in turn.\n");
	fprintf(hlp,"\t            If - is specified, but no ending field number, then all fields\n");
	fprintf(hlp,"\t            starting with field k to the end of the record are printed.\n");
	fprintf(hlp,"\t            Example: Output fields 1-5 with an offset of 1 -- \%f+1-5.\n");
	fprintf(hlp,"\t            n defaults to the length of the field, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the right (Left Justify the field.)\n");
	fprintf(hlp,"\t            k=1...9,a...z to specify up to 35 seperate fields.\n");
	fprintf(hlp,"\t            k can also be a pure decimal number from 1-255, allowing for\n");
	fprintf(hlp,"\t            up to 255 seperate fields.\n");
	fprintf(hlp,"\t            Each field is delimitted by one of the delimiter characters\n");
	fprintf(hlp,"\t            specified by the -D=\"xxx\" argument.  If -D= is omitted, then\n");
	fprintf(hlp,"\t            a space, tab or end of line is used to end a field.\n");
	fprintf(hlp,"\t            If a line ends before Field K is reached, a null string\n");
	fprintf(hlp,"\t            is returned.\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}

	fprintf(hlp,"\t\%[.n]Fk[+m][-l]--Prints n characters Field k starting at position m.\n");
	fprintf(hlp,"\t            If -l is specified, then fields k through l are printed in turn.\n");
	fprintf(hlp,"\t            If - is specified, but no ending field number, then all fields\n");
	fprintf(hlp,"\t            starting with field k to the end of the record are printed.\n");
	fprintf(hlp,"\t            n defaults to the length of the field, m defaults to pos 0.\n");
	fprintf(hlp,"\t            If n is longer than the length of the field, pad with \n");
	fprintf(hlp,"\t            spaces on the left (Right Justify the field.)\n");
	fprintf(hlp,"\t            k=1...9,a...z to specify up to 35 seperate fields.\n");
	fprintf(hlp,"\t            k can also be a pure decimal number from 1-255, allowing for\n");
	fprintf(hlp,"\t            up to 255 seperate fields.\n");
	fprintf(hlp,"\t            Each field is delimitted by one of the delimiter characters\n");
	fprintf(hlp,"\t            specified by the -D=\"xxx\" argument.  If -D= is omitted, then\n");
	fprintf(hlp,"\t            a space, tab or end of line is used to end a field.\n");
	fprintf(hlp,"\t            If a line ends before Field K is reached, a null string\n");
	fprintf(hlp,"\t            is returned.\n");
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}

	fprintf(hlp,"\t\%R          Reads the next line of the file.  \%cC or \%cFn following\n",'%','%');
	fprintf(hlp,"\t            \%R now refers to the new line.\n");
	fprintf(hlp,"\t\%[n]#    -- Prints line number, justified to n positions.\n\t            n defaults to 5.\n");
	fprintf(hlp,"\t\%cn       -- Prints n spaces.  n must be followed by a blank.\n",'%');
	fprintf(hlp,"\t\%cG       -- Prints >.\n",'%');
	fprintf(hlp,"\t\%cL       -- Prints <.\n",'%');
	fprintf(hlp,"\t\%cB       -- Prints |.\n",'%');
	fprintf(hlp,"\t\%cN       -- Prints New Line Character.\n",'%');
	fprintf(hlp,"\t\%cT       -- Prints a tab character.\n",'%');
	fprintf(hlp,"\t\%cQ       -- Prints a double quote (\").\n",'%');
	fprintf(hlp,"\t\%cS       -- Prints a forward slash (/).\n",'%');
	fprintf(hlp,"\t\%cx       -- Prints character x.  Use \%c+ for character + following letter 'c'.\n",'%','%');
	fprintf(hlp,"\t             Use %c- for - anywhere in format.\n",'%');
	fprintf(hlp,"All other characters print as themselves.  Max Linesize is 4096 bytes.\n");
	fprintf(hlp,"If -%c is specified, then use %c instead of %c as shown above.\n",'$','$','%');
	if (hlp == stderr) {
		fprintf(hlp,"\nPlease Press RETURN\n");
		getchar();
	}
	fprintf(hlp,"Example:  Print the first, third and fifth through seventh characters\n");
	fprintf(hlp,"          from each line of CONFIG.SYS, with spaces between each field.\n");
	fprintf(hlp,"          lfmtix CONFIG.SYS  /%c1C %c1C+2 %c3C+4\n\n",'%','%','%');
	fprintf(hlp,"Example:  Print the first 12 positions of the directory display of the\n");
	fprintf(hlp,"          C directory.\n");
	fprintf(hlp,"          ls -l /C | lfmtix  /%c12C \n",'%');
	fprintf(hlp,"Example:  Print only the first 3 positions of field 2 followed by field 15\n");
	fprintf(hlp,"          from file xxx. Use a comma as a field delitter.\n");
	fprintf(hlp,"          lfmtix -D=\",\" xxx /\%c3F2\%cFf\n",'%','%');
	fprintf(hlp,"Example:  Print lines 2,4,6,... of the file xxx\n");
	fprintf(hlp,"          lfmtix xxx /\%cR\%cC\n",'%','%');

	if (hlp != stderr) {
		fclose(hlp);
#ifdef WIN32		
		strcpy(cmd,"NOTEPAD.EXE ");
		strcat(cmd,tmp);
		if (WinExec(cmd,SW_SHOW) <= 31) {
		// An error occurred
			strcpy(cmd,"MORE <");
			strcat(cmd,tmp);
			system(cmd);
		}
#else
		system(tmp);
#endif		
	}	
}


int main(argc,argv)
int argc;
char *argv[];
{
	static char defaultfmt_c[] = "\%#: \%C";
	static char defaultfmt_d[] = "\%#: \%F=";
	char *defaultfmt=defaultfmt_c;
	int i,k,skip=0,j,l;
	char infmt = 0, quiet = 0, inname[_MAX_PATH], inquote;
	char delim[64], temp[64];
	strcpy(delim," \t\n");
	otfi = stdout;
	strcpy(format,defaultfmt);
	strcpy(inname,"STDIN");

	if (argc < 2) {
		Usage();
		exit(0);
	}
#ifdef WIN32	
	_fmode = _O_BINARY;
#endif
	infi = NULL;
	for (i=1;i < argc;++i) {
		switch (argv[i][0]) {
		case '/':
			infmt = 1;
			strcpy(format,argv[i]+1);
			break;
		case '-':
			switch (toupper(argv[i][1])) {
			case 'S':
				skip = atoi(argv[i]+2);
				break;
			case 'D':
				defaultfmt=defaultfmt_d;
				strcpy(format,defaultfmt);
				k = 2;
				inquote = 0;			// Assume no quote
				if (argv[i][k] == '=') // skip =, there for readability
					++k;
				if (argv[i][k] == '\"') { 
					// If the string starts with a quote
					inquote = 1;
					++k;
				}
				if (k < strlen(argv[i]))
					// If a delim has been specified, use it else default
					strcpy(delim,argv[i] + k);
				// remove ending quote if necessary
				if (inquote && delim[strlen(delim)-1] == '\"')
					delim[strlen(delim)-1] = 0;

				// Allow escape characters to be used
				// \t for tab, \n for lf, \r for cr, \x for x

				if (strchr(delim,'\\')) {
					// look for excape characters
					strcpy(temp,delim);
					*delim = 0;

					for (j=0,l=0; j < strlen(temp); ++ j) {
						if (temp[j] == '\\') {
							++j;
							switch (toupper(temp[j])) {
							case 'Q':
								temp[j] = '\"';
								break;
							case 'T':
								temp[j] = 9;
								break;
							case 'N':
								temp[j] = 10;
								break;
							case 'R':
								temp[j] = 13;
								break;
							}
						}
						delim[l++] = temp[j];
					}
					delim[l] = 0;
				}
				// fprintf(stderr,"Delim=\"%s\"\n",delim);
				break;
			case 'Q':
				quiet = 1;
				break;
			case '%':
				cmdchar = '%';
				if (strcmp(format,defaultfmt) == 0) {
					format[0] = '%';
					format[4] = '%';
				}
				defaultfmt[0] = '%';
				defaultfmt[4] = '%';
				break;
			case '$':
				cmdchar = '$';
				if (strcmp(format,defaultfmt) == 0) {
					format[0] = '$';
					format[4] = '$';
				}
				defaultfmt[0] = '$';
				defaultfmt[4] = '$';
				break;
			case 'N':
				usenewline = 0;
				break;
			case 'O':
				otfi = fopen(argv[i]+2,"w");
				if (otfi != NULL)
					fprintf(stderr,"Writing: %s\n",argv[i]+2);
				else {
					fprintf(stderr,"Unable to open output file: %s\nWriting to standard output.\n",argv[i]+2);
					otfi = stdout;
				}
				break;
			case 'C':
				infi = stdin;
				infmt = 2; /* force all other input to be ignored until / */
				break;
			default:
				if (infmt == 1) {
					strcat(format," ");
					strcat(format,argv[i]);
				}
				else 
					fprintf(stderr,"The following text has been ignored:\n%s\n",argv[i]);

			}
			break;
		default:
			switch (infmt) {
			case 0: /* name of input file */
				strcpy(inname,argv[i]);
#ifndef WIN32				
#define stripcm strcasecmp
#endif
				if (stricmp(argv[i],"NUL")) {
					
					infi = fopen(argv[i],"r");
					if (infi == NULL) {
						fprintf(stderr,"Unable to open: %s\n",argv[i]);
						exit(4);
					}
				}
				else 
					infi = NULL;
				infmt = 2;
				break;
			case 1: /* add to format line */
				strcat(format," ");
				strcat(format,argv[i]);
				break;
			default:
				fprintf(stderr,"The following text has been ignored:\n%s\n",argv[i]);
			}						
			break;
		}
	}
	if (!quiet)
		fprintf(stderr,
		"LFMTIX(c) Copyright London Computing, 1987,1988,1992,1994,1997-1999 [%s]\n",__DATE__);



	if (infi == NULL) {
		if (stricmp(inname,"NUL")==0) {
			// Output a single line
			line[0]= 0;
			lineno = 1;
			expand(format,line,delim);
			if (otfi != NULL) 
				fclose(otfi);
			exit(0);
		}
		infi = stdin;
	}
#ifdef WIN32	
	_setmode(_fileno(infi),_O_BINARY);
//	_setmode(_fileno(otfi),_O_BINARY);
#endif
	setvbuf(otfi,outbuff,0,sizeof(outbuff));
	setvbuf(infi,inbuff,0,sizeof(inbuff));
	
	while (skip-- && !done)
		if (fgets(line,sizeof(line),infi) == NULL)
			done = 1;

	while (!done && fgets(line,sizeof(line),infi) != NULL) {
		++lineno;
		expand(format,line,delim);
	}

	if (infi != NULL) fclose(infi);
	if (otfi != NULL) fclose(otfi);
	exit(0);
}
